#
# Copyright (c) 2008-2015 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_form.security module

"""

from pyramid.decorator import reify

from pyams_form.interfaces.form import IFormContextPermissionChecker, IFormSecurityContext

__docformat__ = 'restructuredtext'


def get_checker(request, context=None, view=None, action=''):
    """Get permission checker"""
    registry = request.registry
    if context is None:
        context = request.context
    checker = registry.queryMultiAdapter((context, request, view),
                                         IFormContextPermissionChecker,
                                         name=action) if view is not None else None
    if checker is None:
        checker = registry.queryMultiAdapter((context, request),
                                             IFormContextPermissionChecker,
                                             name=action)
    if checker is None:
        checker = registry.queryAdapter(context, IFormContextPermissionChecker,
                                        name=action)
    return checker


def get_edit_permission(request, context=None, action=''):
    """Get required edit permission"""
    checker = get_checker(request, context, action=action)
    if action and (checker is None):
        checker = get_checker(request, context, action='')
    if checker is not None:
        return checker.edit_permission
    return None


class ProtectedFormObjectMixin:
    """Form object protected by a permission

    A "protected" form is a form on which you apply a permission; the context on which the security
    applies can be provided by an :py:class:`IFormSecurityContext
    <pyams_form.interfaces.IFormSecurityContext>` adapter, or will be extracted for the form
    context itself.

    The permission itself will be provided by an adapter to :py:class:`IFormContextPermissionChecker
    <pyams_form.interfaces.IFormContextPermissionChecker>`
    :

    This class is a form mixin class which should be used for forms protected by a
    security context.
    """

    action_type = ''  # default action

    @reify
    def permission(self):
        """This permission is required to be able to edit the form context"""
        request = self.request  # pylint: disable=no-member
        context = IFormSecurityContext(self, None)
        if context is None:
            context = self.context  # pylint: disable=no-member
        view = getattr(self, '__parent__', None) or \
            getattr(self, 'view', None) or \
            getattr(self, 'table', None)
        checker = get_checker(request, context, view, action=self.action_type)
        if (checker is None) and self.action_type:
            checker = get_checker(request, context, action='')
        if checker is not None:
            return checker.edit_permission
        return None
