#
# Copyright (c) 2008-2015 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_form.widget module

This module provides several custom widgets used inside PyAMS, but also default templates
customized for MyAMS.

This module also provides several decorators, which can be used to define widgets templates
in place of ZCML declarations.

**WARNING**: please note that widgets templates used via z3c.form package use default
"""

import inspect
import json
import locale
import os

import venusian
from pyramid.exceptions import ConfigurationError
from pyramid.interfaces import IRequest
from z3c.form.action import Action
from z3c.form.browser.button import ButtonWidget
from z3c.form.browser.select import SelectWidget
from z3c.form.browser.submit import SubmitWidget
from z3c.form.browser.text import TextWidget
from z3c.form.browser.textarea import TextAreaWidget
from z3c.form.browser.widget import HTMLInputWidget
from z3c.form.button import ButtonAction
from z3c.form.converter import BaseDataConverter, \
    DatetimeDataConverter as BaseDatetimeDataConverter, FormatterValidationError
from z3c.form.interfaces import DISPLAY_MODE, IButtonAction, IDataConverter, IFieldWidget, \
    INPUT_MODE, ITextAreaWidget, IWidgetLayoutTemplate, NO_VALUE
from z3c.form.widget import FieldWidget, Widget, WidgetLayoutFactory, WidgetTemplateFactory
from zope.component import getGlobalSiteManager
from zope.interface import Interface, directlyProvides, implementer, implementer_only
from zope.pagetemplate.interfaces import IPageTemplate
from zope.schema.fieldproperty import FieldProperty
from zope.schema.interfaces import IBytes, IChoice, IDate, IDatetime, IDecimal, IFloat, IInt, \
    IList, ISet, ITime, ITuple

from pyams_form.interfaces.form import IActionWidget, ICloseWidget, IColorWidget, IDateWidget, \
    IDatetimeWidget, IDottedDecimalWidget, IFloatWidget, IFormLayer, IHTMLWidget, IIntegerWidget, \
    IJSONDictWidget, IResetWidget, ISEOTextLineWidget, ISelect2Widget, ITextLineListWidget, \
    ITimeWidget
from pyams_form.schema import IActionButton, ICloseButton, IResetButton
from pyams_skin.interfaces.tinymce import ITinyMCEConfiguration
from pyams_utils.adapter import adapter_config
from pyams_utils.interfaces.data import IObjectData
from pyams_utils.schema import IColorField, IDottedDecimalField, IHTMLField, IJSONDictField, \
    IJSONDictFieldsGetter, ITextLineListField
from pyams_utils.timezone import localgmtime, tztime
from pyams_utils.widget.decimal import DottedDecimalDataConverter


__docformat__ = 'restructuredtext'

from pyams_form import _  # pylint: disable=ungrouped-imports


#
# Widget template configuration
#

class widgettemplate_config:  # pylint: disable=invalid-name
    """Class decorator used to declare a widget template"""

    venusian = venusian  # for testing injection

    def __init__(self, **settings):
        if 'for_' in settings:
            if settings.get('context') is None:
                settings['context'] = settings['for_']
        self.__dict__.update(settings)

    def __call__(self, wrapped):
        settings = self.__dict__.copy()

        def callback(context, name, obj):  # pylint: disable=unused-argument
            template = os.path.join(os.path.dirname(inspect.getfile(inspect.getmodule(obj))),
                                    settings.get('template'))
            if not os.path.isfile(template):
                raise ConfigurationError("No such file", template)

            content_type = settings.get('contentType', 'text/html')
            factory = WidgetTemplateFactory(template, content_type)
            provides = settings.get('provides', IPageTemplate)
            directlyProvides(factory, provides)

            registry = settings.get('registry')
            if registry is None:
                config = context.config.with_package(info.module)  # pylint: disable=no-member
                registry = config.registry
            registry.registerAdapter(factory,
                                     (settings.get('context', Interface),
                                      settings.get('layer', IRequest),
                                      settings.get('view', None),
                                      settings.get('field', None),
                                      settings.get('widget', obj)),
                                     provides,
                                     settings.get('mode', INPUT_MODE))

        info = self.venusian.attach(wrapped, callback, category='pyams_form')

        if info.scope == 'class':  # pylint: disable=no-member
            # if the decorator was attached to a method in a class, or
            # otherwise executed at class scope, we need to set an
            # 'attr' into the settings if one isn't already in there
            if settings.get('attr') is None:
                settings['attr'] = wrapped.__name__

        settings['_info'] = info.codeinfo  # pylint: disable=no-member
        return wrapped


def override_widget_template(widget, **settings):
    """Override template for a given widget

    This function can be used to override a widget template without using ZCML.
    Settings are the same as for @widgettemplate_config decorator.
    """
    stack = inspect.stack()[1]
    template = os.path.join(os.path.dirname(inspect.getfile(inspect.getmodule(stack[0]))),
                            settings.get('template'))
    if not os.path.isfile(template):
        raise ConfigurationError("No such file", template)
    content_type = settings.get('contentType', 'text/html')
    factory = WidgetTemplateFactory(template, content_type)
    provides = settings.get('provides', IPageTemplate)
    directlyProvides(factory, provides)
    registry = getGlobalSiteManager()
    registry.registerAdapter(factory,
                             (settings.get('context', Interface),
                              settings.get('layer', IRequest),
                              settings.get('view', None),
                              settings.get('field', None),
                              widget),
                             provides,
                             settings.get('mode', INPUT_MODE))


class widgetlayout_config:  # pylint: disable=invalid-name
    """Class decorator used to declare a widget layout"""

    venusian = venusian  # for testing injection

    def __init__(self, **settings):
        if 'for_' in settings:
            if settings.get('context') is None:
                settings['context'] = settings['for_']
        self.__dict__.update(settings)

    def __call__(self, wrapped):
        settings = self.__dict__.copy()

        def callback(context, name, obj):  # pylint: disable=unused-argument
            template = os.path.join(os.path.dirname(inspect.getfile(inspect.getmodule(obj))),
                                    settings.get('template'))
            if not os.path.isfile(template):
                raise ConfigurationError("No such file", template)

            content_type = settings.get('contentType', 'text/html')
            factory = WidgetLayoutFactory(template, content_type)
            provides = settings.get('provides', IWidgetLayoutTemplate)
            directlyProvides(factory, provides)

            registry = settings.get('registry')
            if registry is None:
                config = context.config.with_package(info.module)  # pylint: disable=no-member
                registry = config.registry
            registry.registerAdapter(factory,
                                     (settings.get('context', Interface),
                                      settings.get('layer', IRequest),
                                      settings.get('view', None),
                                      settings.get('field', None),
                                      settings.get('widget', obj)),
                                     provides,
                                     settings.get('mode', INPUT_MODE))

        info = self.venusian.attach(wrapped, callback, category='pyams_form')

        if info.scope == 'class':  # pylint: disable=no-member
            # if the decorator was attached to a method in a class, or
            # otherwise executed at class scope, we need to set an
            # 'attr' into the settings if one isn't already in there
            if settings.get('attr') is None:
                settings['attr'] = wrapped.__name__

        settings['_info'] = info.codeinfo  # pylint: disable=no-member
        return wrapped


def override_widget_layout(widget, **settings):
    """Override layout template for a given widget

    This function can be used to override a widget layout template without using ZCML.
    Settings are the same as for @widgetlayout_config decorator.
    """
    stack = inspect.stack()[1]
    template = os.path.join(os.path.dirname(inspect.getfile(inspect.getmodule(stack[0]))),
                            settings.get('template'))
    if not os.path.isfile(template):
        raise ConfigurationError("No such file", template)
    content_type = settings.get('contentType', 'text/html')
    factory = WidgetTemplateFactory(template, content_type)
    provides = settings.get('provides', IWidgetLayoutTemplate)
    directlyProvides(factory, provides)
    registry = getGlobalSiteManager()
    registry.registerAdapter(factory,
                             (settings.get('context', Interface),
                              settings.get('layer', IRequest),
                              settings.get('view', None),
                              settings.get('field', None),
                              widget),
                             provides,
                             settings.get('mode', INPUT_MODE))


#
# Reset button widget and action
#

@widgettemplate_config(mode=INPUT_MODE, template='templates/reset-input.pt', layer=IFormLayer)
@widgettemplate_config(mode=DISPLAY_MODE, template='templates/reset-display.pt', layer=IFormLayer)
@implementer_only(IResetWidget)
class ResetWidget(SubmitWidget):
    """A reset button of a form."""

    klass = u'reset-widget'
    css = u'reset'


@adapter_config(context=(IResetButton, IFormLayer), provides=IFieldWidget)
def ResetFieldWidget(field, request):  # pylint: disable=invalid-name
    """Reset buton field widget factory"""
    reset = FieldWidget(field, ResetWidget(request))
    reset.value = field.title
    return reset


@adapter_config(context=(IFormLayer, IResetButton), provides=IButtonAction)
class ResetButtonAction(ResetWidget, ButtonAction):
    """Reset button action"""

    def __init__(self, request, field):
        # pylint: disable=super-init-not-called
        Action.__init__(self, request, field.title)  # pylint: disable=non-parent-init-called
        ResetWidget.__init__(self, request)  # pylint: disable=non-parent-init-called
        self.field = field


#
# Close button widget and action
#

@widgettemplate_config(mode=INPUT_MODE, template='templates/close-input.pt', layer=IFormLayer)
@widgettemplate_config(mode=DISPLAY_MODE, template='templates/close-display.pt', layer=IFormLayer)
@implementer_only(ICloseWidget)
class CloseWidget(SubmitWidget):
    """A dialog close button"""

    klass = u'close-widget'
    css = u'close'


@adapter_config(context=(ICloseButton, IFormLayer), provides=IFieldWidget)
def CloseFieldWidget(field, request):  # pylint: disable=invalid-name
    """Close button field widget factory"""
    close = FieldWidget(field, CloseWidget(request))
    close.value = field.title
    return close


@adapter_config(context=(IFormLayer, ICloseButton), provides=IButtonAction)
class CloseButtonAction(CloseWidget, ButtonAction):
    """Close button action"""

    def __init__(self, request, field):
        # pylint: disable=super-init-not-called
        Action.__init__(self, request, field.title)  # pylint: disable=non-parent-init-called
        CloseWidget.__init__(self, request)  # pylint: disable=non-parent-init-called
        self.field = field


#
# Action button widget and action
#

@widgettemplate_config(mode=INPUT_MODE, template='templates/action-input.pt', layer=IFormLayer)
@widgettemplate_config(mode=DISPLAY_MODE, template='templates/action-display.pt', layer=IFormLayer)
@implementer_only(IActionWidget)
class ActionWidget(ButtonWidget):
    """A dialog action button"""

    label_css_class = FieldProperty(IActionWidget['label_css_class'])
    click_handler = FieldProperty(IActionWidget['click_handler'])
    url = FieldProperty(IActionWidget['url'])
    modal_target = FieldProperty(IActionWidget['modal_target'])


@adapter_config(context=(IActionButton, IFormLayer), provides=IFieldWidget)
def ActionFieldWidget(field, request):  # pylint: disable=invalid-name
    """Action button field widget factory"""
    action = FieldWidget(field, ActionWidget(request))
    action.value = field.title
    return action


@adapter_config(context=(IFormLayer, IActionButton), provides=IButtonAction)
class ActionButtonAction(ActionWidget, ButtonAction):
    """Action button action"""

    def __init__(self, request, field):
        # pylint: disable=super-init-not-called
        ActionWidget.__init__(self, request)
        Action.__init__(self, request, field.title)  # pylint: disable=non-parent-init-called
        self.field = field
        self.label_css_class = field.label_css_class
        self.click_handler = field.click_handler
        self.url = field.url
        self.modal_target = field.modal_target


#
# Bytes field widget
#

@adapter_config(context=(IBytes, ITextAreaWidget), provides=IDataConverter)
class BytesFieldDataConverter(BaseDataConverter):
    """Bytes field data converter"""

    error_message = _("Invalid UTF-8 encoded data")

    def toWidgetValue(self, value):
        """Convert bytes to string"""
        return value.decode('utf-8') if isinstance(value, bytes) else value

    def toFieldValue(self, value):
        """Convert string to bytes"""
        return value.encode('utf-8') if isinstance(value, str) else value


#
# Integer field widget
#

SIGN_BEFORE_VALUE = (1, 3)
SIGN_AFTER_VALUE = (2, 4)


@adapter_config(context=(IInt, IIntegerWidget), provides=IDataConverter)
class IntegerFieldDataConverter(BaseDataConverter):
    """Integer field data converter"""

    error_message = _("Invalid integer value")

    def toWidgetValue(self, value):
        """Convert integer value to string"""
        if value is self.field.missing_value:
            return ''
        return locale.format_string('%d', value, grouping=True)

    def toFieldValue(self, value):
        """Convert string to integer"""
        if not value:
            return self.field.missing_value
        try:
            return locale.atoi(value)
        except ValueError:
            raise FormatterValidationError(self.error_message, value)


@widgettemplate_config(mode=INPUT_MODE, template='templates/integer-input.pt', layer=IFormLayer)
@implementer_only(IIntegerWidget)
class IntegerWidget(TextWidget):
    """Integer widget"""

    @property
    def validate_rules(self):
        """Add validation rules to JQuery-validate plug-in"""
        conv = locale.localeconv()
        allow_negative = True
        if (self.field.min is not None) and (self.field.min >= 0):
            allow_negative = False
        rules = {
            'pattern': r'{}[\d{}]+{}'.format(
                r'\{}?'.format(conv.get('negative_sign', '-'))
                if (allow_negative and (conv.get('n_sign_posn') in SIGN_BEFORE_VALUE)) else '',
                conv.get('thousands_sep', ''),
                r'\{}?'.format(conv.get('negative_sign', '-'))
                if (allow_negative and (conv.get('n_sign_posn') in SIGN_AFTER_VALUE)) else '')
        }
        return json.dumps(rules)


@adapter_config(context=(IInt, IFormLayer), provides=IFieldWidget)
def IntegerFieldWidget(field, request):  # pylint: disable=invalid-name
    """Integer field widget factory"""
    return FieldWidget(field, IntegerWidget(request))


#
# Float field widget
#

@adapter_config(context=(IFloat, IFloatWidget), provides=IDataConverter)
class FloatFieldDataConverter(BaseDataConverter):
    """Float field data converter"""

    error_message = _("Invalid floating value")

    def toWidgetValue(self, value):
        """Convert float to string"""
        if value is self.field.missing_value:
            return ''
        return locale.format_string('%.{0}f'.format(self.widget.precision), value, grouping=True)

    def toFieldValue(self, value):
        """Convert string to float"""
        if not value:
            return self.field.missing_value
        try:
            return locale.atof(value)
        except ValueError:
            raise FormatterValidationError(self.error_message, value)


@widgettemplate_config(mode=INPUT_MODE, template='templates/float-input.pt', layer=IFormLayer)
@implementer_only(IFloatWidget)
class FloatWidget(TextWidget):
    """Float widget"""

    precision = FieldProperty(IFloatWidget['precision'])

    @property
    def validate_rules(self):
        """Add validation rules to JQuery-validate plug-in"""
        conv = locale.localeconv()
        allow_negative = True
        if (self.field.min is not None) and (self.field.min >= 0):
            allow_negative = False
        rules = {
            'pattern': r'{}[\d{}]+{}?\d*{}'.format(
                r'\{}?'.format(conv.get('negative_sign', '-'))
                if (allow_negative and (conv.get('n_sign_posn') in SIGN_BEFORE_VALUE)) else '',
                conv.get('thousands_sep', ''),
                conv.get('decimal_point', r'\.'),
                r'\{}?'.format(conv.get('negative_sign', '-'))
                if (allow_negative and (conv.get('n_sign_posn') in SIGN_AFTER_VALUE)) else '')
        }
        return json.dumps(rules)


@adapter_config(context=(IFloat, IFormLayer), provides=IFieldWidget)
@adapter_config(context=(IDecimal, IFormLayer), provides=IFieldWidget)
def FloatFieldWidget(field, request):  # pylint: disable=invalid-name
    """Float field widget factory"""
    return FieldWidget(field, FloatWidget(request))


@widgettemplate_config(mode=INPUT_MODE, template='templates/decimal-input.pt', layer=IFormLayer)
@implementer_only(IDottedDecimalWidget)
class DottedDecimalWidget(FloatWidget):
    """Dotted decimal field widget"""

    @property
    def validate_rules(self):
        """Add validation rules to JQuery-validate plug-in"""
        conv = locale.localeconv()
        allow_negative = True
        if (self.field.min is not None) and (self.field.min >= 0):
            allow_negative = False
        rules = {
            'pattern': r'{}[\d{}]+{}?\d*{}'.format(
                r'\{}?'.format(conv.get('negative_sign', '-'))
                if (allow_negative and (conv.get('n_sign_posn') in SIGN_BEFORE_VALUE)) else '',
                '',  # no thousands separator
                r'\.',  # dot as decimal separator !!
                r'\{}?'.format(conv.get('negative_sign', '-'))
                if (allow_negative and (conv.get('n_sign_posn') in SIGN_AFTER_VALUE)) else '')
        }
        return json.dumps(rules)

    @property
    def validate_messages(self):
        """Get validation message for pattern rule"""
        return json.dumps({
            'pattern': self.request.localizer.translate(DottedDecimalDataConverter.errorMessage)
        })


@adapter_config(context=(IDottedDecimalField, IFormLayer), provides=IFieldWidget)
def DottedDecimalFieldWidget(field, request):  # pylint: disable=invalid-name
    """Dotted decimal field widget factory"""
    return FieldWidget(field, DottedDecimalWidget(request))


#
# Date widget
#

@widgettemplate_config(mode=INPUT_MODE, template='templates/date-input.pt', layer=IFormLayer)
@implementer_only(IDateWidget)
class DateWidget(TextWidget):
    """Date widget"""


@adapter_config(context=(IDate, IFormLayer), provides=IFieldWidget)
def DateFieldWidget(field, request):  # pylint: disable=invalid-name
    """Date field widget factory"""
    return FieldWidget(field, DateWidget(request))


#
# Datetime widget
#

@adapter_config(context=(IDatetime, IDatetimeWidget), provides=IDataConverter)
class DatetimeDataConverter(BaseDatetimeDataConverter):
    """Datetime field data converter"""

    def toWidgetValue(self, value):
        """Convert datetime to string"""
        value = tztime(value)
        return super(DatetimeDataConverter, self).toWidgetValue(value)

    def toFieldValue(self, value):
        """Convert string to local datetime"""
        value = super(DatetimeDataConverter, self).toFieldValue(value)
        return localgmtime(value)


@widgettemplate_config(mode=INPUT_MODE, template='templates/datetime-input.pt', layer=IFormLayer)
@implementer_only(IDatetimeWidget)
class DatetimeWidget(TextWidget):
    """Datetime widget"""


@adapter_config(context=(IDatetime, IFormLayer), provides=IFieldWidget)
def DatetimeFieldWidget(field, request):  # pylint: disable=invalid-name
    """Datetime field widget factory"""
    return FieldWidget(field, DatetimeWidget(request))


#
# Time widget
#

@widgettemplate_config(mode=INPUT_MODE, template='templates/time-input.pt', layer=IFormLayer)
@implementer_only(ITimeWidget)
class TimeWidget(TextWidget):
    """Time widget"""


@adapter_config(context=(ITime, IFormLayer), provides=IFieldWidget)
def TimeFieldWidget(field, request):  # pylint: disable=invalid-name
    """Time field widget factory"""
    return FieldWidget(field, TimeWidget(request))


#
# Color widget
#

@adapter_config(context=(IColorField, IColorWidget), provides=IDataConverter)
class ColorDataConverter(BaseDataConverter):
    """Color field data converter"""

    def toWidgetValue(self, value):
        """Convert color to string"""
        value = super(ColorDataConverter, self).toWidgetValue(value)
        if value:
            value = '#' + value
        return value

    def toFieldValue(self, value):
        """Convert string to color value"""
        if value and value.startswith('#'):
            value = value[1:]
        return super(ColorDataConverter, self).toFieldValue(value)


@widgettemplate_config(mode=INPUT_MODE, template='templates/color-input.pt', layer=IFormLayer)
@implementer_only(IColorWidget)
class ColorWidget(TextWidget):
    """Color widget"""


@adapter_config(context=(IColorField, IFormLayer), provides=IFieldWidget)
def ColorFieldWidget(field, request):  # pylint: disable=invalid-name
    """Color field widget factory"""
    return FieldWidget(field, ColorWidget(request))


#
# HTML widget
#

@widgettemplate_config(mode=INPUT_MODE, template='templates/html-input.pt', layer=IFormLayer)
@widgettemplate_config(mode=DISPLAY_MODE, template='templates/html-display.pt', layer=IFormLayer)
@implementer_only(IHTMLWidget)
class HTMLWidget(TextAreaWidget):
    """HTML editor widget"""

    @property
    def editor_data(self):
        """Get editor data"""
        registry = self.request.registry
        config = registry.queryMultiAdapter((self.form, self.request), ITinyMCEConfiguration,
                                            name=self.name)
        if (config is None) and hasattr(self, 'basename'):  # I18n widget
            config = registry.queryMultiAdapter((self.form, self.request), ITinyMCEConfiguration,
                                                name=self.basename)  # pylint: disable=no-member
        if config is None:
            config = registry.queryMultiAdapter((self.form, self.request), ITinyMCEConfiguration)
        if config is None:
            config = getattr(self, 'editor_config_data', None)
            if config is not None:
                return json.dumps(config)
            return None
        return json.dumps(config.configuration)


@adapter_config(context=(IHTMLField, IFormLayer), provides=IFieldWidget)
def HTMLFieldWidget(field, request):  # pylint: disable=invalid-name
    """HTML editor field widget factory"""
    return FieldWidget(field, HTMLWidget(request))


#
# JSON dict widget
#

@adapter_config(context=(IJSONDictField, IJSONDictWidget), provides=IDataConverter)
class JSONDictDataConverter(BaseDataConverter):
    """JSON dict data converter"""

    def toWidgetValue(self, value):
        """Convert JSON dict to string"""
        return value

    def toFieldValue(self, value):
        """Load JSON from string"""
        return json.loads(value)


@widgettemplate_config(mode='display', template='templates/json-dict-display.pt', layer=IFormLayer)
@implementer_only(IJSONDictWidget)
class JSONDictWidget(HTMLInputWidget, Widget):
    """JSON dict field widget"""

    def get_fields(self):
        """Get fields from JSON mapping"""
        form = self.__parent__.form
        getter = self.request.registry.queryMultiAdapter((form.getContent(), self.request, form),
                                                         IJSONDictFieldsGetter)
        if getter is not None:
            yield from getter.get_fields(self.value)
        else:
            yield from map(lambda item: (item[0], item[0], item[1]), self.value.items())


@adapter_config(context=(IJSONDictField, IFormLayer), provides=IFieldWidget)
def JSONDictFieldWidget(field, request):  # pylint: disable=invalid-name
    """JSON dict field widget factory"""
    return FieldWidget(field, JSONDictWidget(request))


#
# Select2 widget
#

@widgettemplate_config(mode=INPUT_MODE, template='templates/select-input.pt',
                       layer=IFormLayer)
@widgettemplate_config(mode=DISPLAY_MODE, template='templates/select-display.pt',
                       layer=IFormLayer)
@implementer_only(ISelect2Widget)
class Select2Widget(SelectWidget):
    """Select2 widget"""

    noValueMessage = _("(no selected value)")

    def get_content(self, entry):
        """Get translated entry content"""
        translate = self.request.localizer.translate
        return translate(entry['content'])


@adapter_config(context=(IChoice, IFormLayer), provides=IFieldWidget)
def ChoiceFieldWidget(field, request):  # pylint: disable=invalid-name
    """Choice field widget factory"""
    return FieldWidget(field, Select2Widget(request))


@widgettemplate_config(mode=INPUT_MODE, template='templates/hidden-select-input.pt',
                       layer=IFormLayer)
@widgettemplate_config(mode=DISPLAY_MODE, template='templates/hidden-select-display.pt',
                       layer=IFormLayer)
@implementer(IObjectData)
class HiddenSelect2Widget(Select2Widget):
    """Hidden select2 widget

    Widget values are provided through IObjectData
    """

    def extract(self, default=NO_VALUE):
        """Extract terms from input value"""
        if self.name not in self.request and self.name + '-empty-marker' in self.request:
            return []
        value = self.request.get(self.name, default)
        if value != default:
            if not isinstance(value, (set, tuple, list)):
                value = set(value.split('|')) if value else set()
            # do some kind of validation, at least only use existing values
            for token in value:
                if token == self.noValueToken:
                    continue
                try:
                    self.terms.getTermByToken(token)
                except LookupError:
                    return default
        return value

    @property
    def values(self):
        """Convert values list to string"""
        return '|'.join(self.value or '')

    @property
    def values_map(self):
        """Create JSON mapping from selected values"""
        result = {}
        terms = self.terms
        # pylint: disable=expression-not-assigned
        [result.update({value: terms.getTermByToken(value).title}) for value in self.value or ()]
        return json.dumps(result)


def HiddenSelect2FieldWidget(field, request):  # pylint: disable=invalid-name
    """Hidden Select2 field widget factory"""
    return FieldWidget(field, HiddenSelect2Widget(request))


@adapter_config(context=(IList, HiddenSelect2Widget), provides=IDataConverter)
@adapter_config(context=(ITuple, HiddenSelect2Widget), provides=IDataConverter)
@adapter_config(context=(ISet, HiddenSelect2Widget), provides=IDataConverter)
class HiddenSelect2DataConverter(BaseDataConverter):
    """Hidden select2 data converter"""

    def toWidgetValue(self, value):
        """Convert selection to string"""
        return value or ()

    def toFieldValue(self, value):
        """Create set from selected values"""
        return set(value or ())


#
# TextLineList widget
#

@adapter_config(context=(ITextLineListField, ITextLineListWidget), provides=IDataConverter)
class TextLineListFieldDataConverter(BaseDataConverter):
    """Text line list field data converter"""

    def toWidgetValue(self, value):
        """Convert values list to string"""
        return '|'.join(value or [])

    def toFieldValue(self, value):
        """Create list from string"""
        return value.split('|') if value else None


@widgettemplate_config(mode=INPUT_MODE, template='templates/textlinelist-input.pt',
                       layer=IFormLayer)
@widgettemplate_config(mode=DISPLAY_MODE, template='templates/textlinelist-display.pt',
                       layer=IFormLayer)
@implementer_only(ITextLineListWidget)
class TextLineListWidget(TextWidget):
    """Text line list widget"""

    @property
    def tags(self):
        """Create string from input value"""
        return json.dumps((self.value or '').split('|'))


@adapter_config(context=(ITextLineListField, IFormLayer), provides=IFieldWidget)
def TextLineListFieldWidget(field, request):  # pylint: disable=invalid-name
    """Textlines list field widget"""
    return FieldWidget(field, TextLineListWidget(request))


#
# TextLine widget with SEO length status
#

@widgettemplate_config(mode=INPUT_MODE, template='templates/seo-textline-input.pt',
                       layer=IFormLayer)
@implementer_only(ISEOTextLineWidget)
class SEOTextLineWidget(TextWidget):
    """SEO textline widget"""

    @property
    def length(self):
        """Get current length of text"""
        return len(self.value or '')

    @property
    def status(self):
        """Get widget status based on text length; a "good" length is between 40 and 66
        characters
        """
        status = 'success'
        length = self.length
        if length < 20 or length > 80:
            status = 'danger'
        elif length < 40 or length > 66:
            status = 'warning'
        return status


def SEOTextLineFieldWidget(field, request):  # pylint: disable=invalid-name
    """SEO textline field widget factory"""
    return FieldWidget(field, SEOTextLineWidget(request))
